﻿
//====================================================
// Описание работы классов и методов исходника на:
// https://www.interestprograms.ru
// Исходные коды программ и игр
// Автор исходных кодов Клочков Павел
//====================================================

namespace WinFormsTicTacToe.GameClasses
{
    internal class GameCanvas : Control
    {
        // Количество ячеек в ряду.
        private readonly int _numcell;
        // Размер ячейки.
        private int _sizecell;
        // Массив фигурок на поле, индексы - это кординаты.
        private readonly Figures[,] _figures;
        // Тип фигуры для игрового поля текущего игрока.
        public bool _cross = true;
        // Флаг рисования линии полностью заполненной
        // фигурками одного типа.
        private bool _drawFullLine;

        // Линия для визуальной акцентированности законченного ряда.
        // Содержит начальную и конечную координату линии.
        private readonly Line _line = new();

        // Визуальность активного и неактивного состояния
        private bool _activeState = false;
        public bool ActiveState
        {
            get { return _activeState; }
            set
            {
                _activeState = value;
                Invalidate();
            }
        }

        // Делегаты событий.
        public event EventHandler<PlayerMoveEventArgs>? EventPlayerMove;
        public event EventHandler? EventGameVictory;
        public event EventHandler? EventGameDraw;

        public GameCanvas(int numcell = 3)
        {
            // Устраняет мерцание графики.
            this.DoubleBuffered = true;

            // Определение количества ячеек.
            _numcell = numcell;

            // Создание массива расположения фигур на поле игры.
            _figures = new Figures[_numcell, _numcell];

            for (int x = 0; x < _numcell; x++)
            {
                for (int y = 0; y < _numcell; y++)
                {
                    _figures[x, y] = new();
                }
            }
        }


        protected override void OnResize(EventArgs e)
        {
            base.OnResize(e);

            // Определение размера ячеек после изменения
            // размера поля игры.
            _sizecell = Width / _numcell;
        }

        protected override void OnPaint(PaintEventArgs e)
        {
            base.OnPaint(e);

            // Цвета активного состояния игрового поля.
            // Цвет земли поля.
            Brush rcBrush = Constants.rcBrush;
            // Цвет линий сетки.
            Brush cellsBrush = Constants.cellsBrush;
            // Цвет фигур.
            Brush figureBrush = Constants.cellsBrush;
            // Цвет линии визуально обозначающую заполненность
            // одинаковыми фигурами.
            Brush lineBrush = Constants.lineBrush;


            // Цвета неактивного состояния игрового поля.
            if (ActiveState == false)
            {
                rcBrush = Constants.rcBrushDeactive;
                figureBrush = Constants.cellsBrushDeactive;
                cellsBrush = Constants.cellsBrushDeactive;
            }

            // Закраска заднего фона поля.
            Graphics g = e.Graphics;
            g.FillRectangle(rcBrush, 0, 0, Width, Height);

            // Рисование сетки ячеек
            for (int x = 1; x < _numcell; x++)
            {
                g.DrawLine(new Pen(cellsBrush), 0, _sizecell * x, Width, _sizecell * x);
                g.DrawLine(new Pen(cellsBrush), _sizecell * x, 0, _sizecell * x, Height);
            }

            // Рисование фигур
            for (int x = 0; x < _numcell; x++)
            {
                for (int y = 0; y < _numcell; y++)
                {
                    Point origin = new(_sizecell * x + _sizecell / 2, _sizecell * y + _sizecell / 2);
                    _figures[x, y].Draw(g, origin, _sizecell, figureBrush);
                }
            }

            // Рисование линии заполненной одинаковыми фигурами
            // в случае победы одного из игроков.
            if (_drawFullLine == true)
            {
                g.DrawLine(new Pen(lineBrush, 14), _line.p1.X, _line.p1.Y, _line.p2.X, _line.p2.Y);
            }
        }


        protected override void OnMouseDown(MouseEventArgs e)
        {
            base.OnMouseDown(e);

            // Событие MouseDown обрабатывается только при активном поле игры. 
            // В неактивном состоянии событие хода игрока и 
            // события окончания игры не генерируются.
            if (ActiveState == false) return;

            // Вычисляем ячейку над которой пользователь щелкнул кнопкой мыши.
            for (int x = 0; x < _numcell; x++)
            {
                for (int y = 0; y < _numcell; y++)
                {
                    if (e.X > x * _sizecell &&
                        e.X < (x + 1) * _sizecell &&
                        e.Y > y * _sizecell &&
                        e.Y < (y + 1) * _sizecell)
                    {
                        // Координата в номерах ячейки вычислена.
                        PlayerMoveEventArgs args = new()
                        {
                            X = (byte)x,
                            Y = (byte)y
                        };

                        // Проверяем: есть ли в данной клетке уже нарисованная фигура,
                        // если есть - запрещаем генерирование событий.
                        Figures f = _figures[x, y];
                        if (f.type != 0)
                        {
                            // Звук ошибки.
                            Console.Beep(500, 200);
                            return;
                        }

                        // Генерация события хода пользователя с координатами расположения ячейки хода.
                        // Используется в главной форме.
                        EventPlayerMove?.Invoke(this, args);

                        // После каждого хода проверяем игру на окончание.
                        // Если какая-либо линия (ряд, столбец, одна из больших диагоналей)
                        // заполнена однообразными фигурами, генерируется событие окончания игры
                        // с победителем и проигравшим.
                        if (SameShapesFullLine() == true)
                        {
                            // Генерация события окончания игры.
                            EventGameVictory?.Invoke(this, new EventArgs());
                        }

                        // Проверка на окончание игры с результатом ничья.
                        if (CheckDraw() == true)
                        {
                            // Генерация события окончания игры.
                            EventGameDraw?.Invoke(this, new EventArgs());
                        }

                        return;
                    }
                }
            }
        }

        /// <summary>
        /// Команда на рисование фигурки на игровом поле по координатам ячеек.
        /// </summary>
        /// <param name="x">порядковый номер ячейки по горизонтали</param>
        /// <param name="y">порядковый номер ячейки по вертикали</param>
        /// <param name="type">вид фигурки(крестик или нолик)</param>
        public void DoDraw(int x, int y, int type)
        {
            if (ActiveState == true)
            {
                Figures f = _figures[x, y];

                if (f.type == 0)
                {
                    f.type = type;

                    // При строки, столбца, диагонали однотипными фигурами
                    // рисуется линия визуально обозначающая заполнение.
                    // Вызов метода в данном месте только для рисования линии заполненности.
                    SameShapesFullLine();

                    Invalidate();
                }
            }
        }



        /// <summary>
        /// Очистка поля игры
        /// </summary>
        public void Clear()
        {
            // Стираем линию заполненной фигурами одного типа.
            _drawFullLine = false;

            // Активируем игровое поле
            ActiveState = true;

            // Вид фигуры поля неопределён - снова пустое поле.
            foreach (Figures f in _figures)
            {
                f.type = 0;
            }

            Invalidate();
        }

        /// <summary>
        /// Проверка заполнения полной линии одинаковыми фигурами 
        /// </summary>
        /// <returns>true - линия заполнена</returns>
        private bool SameShapesFullLine()
        {
            // Проверка вертикальных линий
            for (int x = 0; x < _numcell; x++)
            {
                // Начальная координата заполненной линии.
                // Линия отмечает заполненный ряд.
                _line.p1 = new Point((_sizecell * x) + _sizecell / 2, 0);

                // Счётчик совпадающих фигур в исследуемом ряде,
                // начинается с первой фигуры.
                int count = 1;
                for (int y = 0; y < _numcell - 1; y++)
                {
                    Figures f1 = _figures[x, y];
                    Figures f2 = _figures[x, y + 1];

                    // Если фигура есть в ячейке.
                    if (f1.type != 0)
                    {
                        // если тип первой совпадает со второй,
                        // если вторая совпадает с третьей и т.д.,
                        // то линия заполнена одинаковыми фигурами.
                        if (f1.type == f2.type)
                        {
                            count++;
                        }
                    }
                }

                // Конечная координата заполненной линии.
                // Линия показывает заполненный ряд.
                _line.p2 = new Point((_sizecell * x) + _sizecell / 2, (_numcell * _sizecell));

                // Вывод результата в виде возврата: истина - столбец заполнен фигурами одного вида, ложь - всё  остальное.
                if (ShowGameOver(count, _numcell) == true) return true;
            }

            // Проверка горизонтальных линий
            for (int y = 0; y < _numcell; y++)
            {
                // Начальная координата заполненной линии.
                // Линия отмечает заполненный ряд.
                _line.p1 = new Point(0, (y * _sizecell) + _sizecell / 2);

                // Счётчик совпадающих фигур в исследуемом ряде,
                // начинается с первой фигуры.
                int count = 1;
                for (int x = 0; x < _numcell - 1; x++)
                {
                    Figures f1 = _figures[x, y];
                    Figures f2 = _figures[x + 1, y];

                    // Если фигура есть в ячейке
                    if (f1.type != 0)
                    {
                        // Последовательная проверка совпадения видов фигур в ячейках.
                        if (f1.type == f2.type)
                        {
                            count++;
                        }
                    }
                }

                // Конечная координата заполненной линии.
                // Линия показывает заполненный ряд.
                _line.p2 = new Point(_numcell * _sizecell, (y * _sizecell) + _sizecell / 2);


                // Вывод результата в виде возврата: истина - ряд заполнен фигурами одного вида, ложь - всё  остальное.
                if (ShowGameOver(count, _numcell) == true) return true;
            }


            //----
            // Проверка центральной диагонали,
            // левый-верхний, правый-нижний.
            int countfigure = 1;
            for (int i = 0; i < _numcell - 1; i++)
            {
                Figures f = _figures[i, i];
                Figures f2 = _figures[i + 1, i + 1];

                // Если фигура выбрана,
                if (f.type != 0)
                {
                    // если тип первой совпадает со второй,
                    // если вторая совпадает с третьей,
                    // то в ряд заполнен одинаковыми фигурами.
                    if (f.type == f2.type)
                    {
                        countfigure++;
                    }
                }
            }

            // Вывод результата в виде возврата: истина - диагональ заполнена фигурами одного вида, ложь - всё  остальное.
            if (ShowGameOver(countfigure, _numcell) == true)
            {
                // Визуальная линия выделения диагонали однотипных фигур.
                _line.p1 = new Point(0, 0);
                _line.p2 = new Point(_numcell * _sizecell, _numcell * _sizecell);

                return true;
            }

            //----

            //----
            // Проверка центральной диагонали,
            // правый-верхний, левый-нижний.
            int countfigure2 = 1;
            for (int y = 0; y < _numcell - 1; y++)
            {
                int x = (_numcell - 1) - y;

                Figures f = _figures[x, y];
                Figures f2 = _figures[x - 1, y + 1];

                // Если фигура выбрана,
                if (f.type != 0)
                {
                    // если тип первой совпадает со второй,
                    // если вторая совпадает с третьей,
                    // то в ряд заполнен одинаковыми фигурами.
                    if (f.type == f2.type)
                    {
                        countfigure2++;
                    }
                }
            }

            // Вывод результата в виде возврата: истина - диагональ заполнена фигурами одного вида, ложь - всё  остальное.
            if (ShowGameOver(countfigure2, _numcell) == true)
            {
                _line.p1 = new Point(_numcell * _sizecell, 0);
                _line.p2 = new Point(0, _numcell * _sizecell);

                return true;
            }


            // Внутренняя функция сравнения количества однородных фигур
            // с количеством ячеек в иследоваемой протяженности(ряд, столбец, диагональ).
            bool ShowGameOver(int check, int numcell)
            {
                // Если ряд запонен одинаковыми фигурами, можно рисовать 
                // линию вдоль фигур.
                if (check == numcell)
                {
                    // Разрешение на рисование линии заполненого ряда.
                    _drawFullLine = true;

                    return true;
                }

                return false;
            }


            return false;
        }

        /// <summary>
        /// Проверка состояния ничья 
        /// Проверка присутствия разных фигур во всех рядах, столбцах, больших диагоналях. 
        /// </summary>
        /// <returns></returns>
        private bool CheckDraw()
        {
            // Счетчики заполненных линий для состояния ничья.
            // Счетчик для горизонтальных линий
            int linevertical = 0;
            // Счетчик для вертикальных линий.
            int linehorizontal = 0;


            // Проверка вертикальных и горизонтальных линий.
            for (int col = 0; col < _numcell; col++)
            {
                // Счётчики соответствующих фигур
                var vertical = (count1: 0, count2: 0);
                var horizontal = (count1: 0, count2: 0);

                for (int row = 0; row < _numcell; row++)
                {
                    // Проверка вертикальных линий
                    Figures fv = _figures[col, row];
                    if (fv.type == 1) vertical.count1++;
                    if (fv.type == 2) vertical.count2++;

                    // Проверка горизонтальных линий 
                    Figures fh = _figures[row, col];
                    if (fh.type == 1) horizontal.count1++;
                    if (fh.type == 2) horizontal.count2++;
                }

                // Если на линии располагаются две разные фигуры,
                // на такой линии победить нельзя.
                if (vertical.count1 > 0 && vertical.count2 > 0)
                {
                    linevertical++;
                }

                // Если на линии располагаются две разные фигуры,
                // на такой линии победить нельзя.
                if (horizontal.count1 > 0 && horizontal.count2 > 0)
                {
                    linehorizontal++;
                }
            }

            // Счётчики соответствующих фигур
            var diagonal1 = (count1: 0, count2: 0);
            var diagonal2 = (count1: 0, count2: 0);
            for (int cell = 0; cell < _numcell; cell++)
            {
                // Проверка диагонали левый-верхний \ правый-нижний. 
                Figures f1 = _figures[cell, cell];
                if (f1.type == 1) diagonal1.count1++;
                if (f1.type == 2) diagonal1.count2++;

                // Проверка диагонали правый-верхний / левый-нижний. 
                Figures f2 = _figures[(_numcell - 1) - cell, cell];
                if (f2.type == 1) diagonal2.count1++;
                if (f2.type == 2) diagonal2.count2++;
            }

            // Обработка полученных данных
            // Если на всех линиях есть фигуры разного типа,
            // игру продолжать бесмысленно - ничья.
            if (linevertical == _numcell && linehorizontal == _numcell &&
                diagonal1.count1 > 0 && diagonal1.count2 > 0 &&
                diagonal2.count1 > 0 && diagonal2.count2 > 0)
            {
                // MessageBox.Show("Ничья, победила дружба!");
                return true;
            }


            return false;
        }

    }

    public class PlayerMoveEventArgs : EventArgs
    {
        // Координаты хода игрока
        public byte X;
        public byte Y;
    }
}
